/*
 *   This is dvips, a freely redistributable PostScript driver
 *   for dvi files.  It is (C) Copyright 1987 by Tomas Rokicki.
 *   You may modify and use this program to your heart's content,
 *   so long as you send modifications to Tomas Rokicki.  It can
 *   be included in any distribution, commercial or otherwise, so
 *   long as the banner string defined in structures.h is not
 *   modified (except for the version number) and this banner is
 *   printed on program invocation, or can be printed on program
 *   invocation with the -? option.
 */
/*
 *   dospecial of dvisw package.  (C) 1986 Radical Eye Software.
 *   This routine handles special commands.  Currently the only
 *   special command handled is the landscape special.
 */
#include "structures.h"
#include <string.h>
/*
 *   These are the external routines called:
 */
/**/
#ifdef TPIC
extern void setPenSize();
extern void flushPath();
extern void flushDashed();
extern void flushDashed();
extern void addPath();
extern void arc();
extern void flushSpline();
extern void shadeLast();
extern void whitenLast();
extern void blackenLast();
#endif
extern shalfword dvibyte() ;
extern void add_header() ;
extern void hvpos() ;
extern void figcopyfile() ;
extern char *malloc() ;
extern void cmdout() ;
extern void error() ;
extern shalfword linepos;
extern Boolean prescan;
extern Boolean landscape, orientationset ;
extern char *paperfmt ;
extern char *nextstring;
extern char *maxstring;
extern char *oname;
extern FILE *bitfile;
extern int quiet;
#ifdef DEBUG
extern integer debug_flag;
#endif
/*
 *   Now our routine.  We get the number of bytes specified and place them
 *   into the string buffer, and then parse it for alpha.  Interpret a 
 *   \special command, made up of keyword=value pairs.
 */

#define NKEYS    (sizeof(KeyTab)/sizeof(KeyTab[0]))

KeyDesc KeyTab[] = {{"psfile",  String},
                    {"ifffile", String},
                    {"tekfile", String},
                    {"hsize",   Number},
                    {"vsize",   Number},
                    {"hoffset", Number},
                    {"voffset", Number},
                    {"hscale",  Number},
                    {"vscale",  Number},
                    {"angle",   Number},
                    {"llx", Number},
                    {"lly", Number},
                    {"urx", Number},
                    {"ury", Number},
                    {"rwi", Number}};

char  *GetKeyStr();

struct bangspecial {
   struct bangspecial *next ;
   char actualstuff[4] ;
} *bangspecials = NULL ;

static trytobreakout(p)
register char *p ;
{
   register int i ;
   register int instring = 0 ;
   int lastc = 0 ;

   i = 0 ;
   while (*p) {
      if (i > 65 && *p == ' ' && instring == 0) {
         (void)putc(10, bitfile) ;
         i = 0 ;
      } else {
         (void)putc(*p, bitfile) ;
         i++ ;
      }
      if (*p == '(' && lastc != '\\')
         instring = 1 ;
      else if (*p == ')' && lastc != '\\')
         instring = 0 ;
      lastc = *p ;
      p++ ;
   }
}

static dobs(q)
register struct bangspecial *q ;
{
   if (q) {
      dobs(q->next) ;
      trytobreakout(q->actualstuff) ;
   }
}

void
outbangspecials() {
   if (bangspecials) {
      cmdout("TeXDict") ;
      cmdout("begin") ;
      cmdout("@defspecial\n") ;
      dobs(bangspecials) ;
      cmdout("\n@fedspecial") ;
      cmdout("end") ;
   }
}

#define TRUE (1)
#define FALSE (0)

void dospecial(numbytes)
integer numbytes ;
{
   register char *p = nextstring ;
   register int i = 0 ;
   int j ;
   Boolean ps = TRUE ;
   char spbuf[100] ; 
   char fbuf[100] ; 
   char *sf = NULL ;
   char *f = NULL ;
   KeyWord k ;

   if (nextstring + i > maxstring)
      error("! out of string space in dospecial") ;
   for (i=numbytes; i>0; i--)
      *p++ = (char)dvibyte() ;
   *p = 0 ;
   p = nextstring ;
   while (*p <= ' ' && *p)
      p++ ;
#ifdef DEBUG
   if (dd(D_SPECIAL))
      (void)fprintf(stderr, "Processing special: %s\n", p) ;
#endif

   if (strncmp(p, "landscape", 9)==0) {
      if (prescan)
         if (orientationset)
            error("landscape special too late!") ;
         else {
            landscape = 1 ;
            paperfmt = "@landscape" ;
         }
      return ;
   }
   if (strncmp(p, "header", 6)==0) {
      char *q ;
      p += 6 ;
      while ((*p <= ' ' || *p == '=' || *p == ')') && *p != 0)
         p++ ;
      q = p ;
      p = p + strlen(p) - 1 ;
      if (*p == ')') {
         p-- ;
         while (*p <= ' ')
            p-- ;
         p[1] = 0 ;
      }
      add_header(q) ;
   }
   else if (strncmp(p, "ps:", 3)==0) {
      if (! prescan) {
        hvpos() ;
	if (strncmp(p, "ps::[begin]", 11) == 0)
	   cmdout(&p[11]);
	else if (strncmp(p, "ps::[end]", 9) == 0)
	   cmdout(&p[9]);
	else if (strncmp(p, "ps: plotfile ", 13) == 0) {
           char *sfp ;
	   p += 13;
	   for (sfp = p; *sfp && *sfp != ' '; sfp++) ;
	   *sfp = '\0';
	   figcopyfile (p);
	} else if (strncmp(p, "ps::", 4) == 0)
	   cmdout(&p[4]);
	else
	   cmdout(&p[3]);
	return;
      }
   }
#ifdef TPIC
   else if (strncmp(p, "pn ", 3) == 0) {if (!prescan) setPenSize(p+2);}
   else if (strncmp(p, "fp",2) == 0) {if (!prescan) flushPath();}
   else if (strncmp(p, "da ", 3) == 0) {if (!prescan) flushDashed(p+2, 0);}
   else if (strncmp(p, "dt ", 3) == 0) {if (!prescan) flushDashed(p+2, 1);}
   else if (strncmp(p, "pa ", 3) == 0) {if (!prescan) addPath(p+2);}
   else if (strncmp(p, "ar ", 3) == 0) {if (!prescan) arc(p+2);}
   else if (strncmp(p, "sp", 2) == 0) {if (!prescan) flushSpline();}
   else if (strncmp(p, "sh", 2) == 0) {if (!prescan) shadeLast();}
   else if (strncmp(p, "wh", 2) == 0) {if (!prescan) whitenLast();}
   else if (strncmp(p, "bk", 2) == 0) {if (!prescan) blackenLast();}
   else if (strncmp(p, "tx ", 3) == 0) {if (!prescan) SetShade(p+3);}
#endif
   else {
      if (prescan && *p == '!') {
         struct bangspecial *q ;
   
         p++ ;
         q = (struct bangspecial *)malloc((unsigned)
                            (sizeof(struct bangspecial) + strlen(p))) ;
         if (q == NULL)
            error("! out of memory in dospecial") ;
         (void)strcpy(q->actualstuff, p) ;
         q->next = bangspecials ;
         bangspecials = q ;
         return ;
      }
   
      if (! prescan) {
   
         if (*p == '"') {
            hvpos();
            cmdout("@beginspecial") ;
            cmdout("@setspecial\n") ;
            trytobreakout(p+1) ;
            cmdout("\n@endspecial") ;
            return ;
         } else if (*p == '!')
            return ;
      
         spbuf[0] = '\0';
   
         hvpos();
   
         cmdout("@beginspecial\n");
   
         /* get all keyword-value pairs; for compatibility, 
         single words are taken as file names */

      while( (p=GetKeyStr(p,&k)) != NULL ) {
         if( k.vt == None && access(k.Key,0) == 0) {
            if( sf ) {
               (void)fprintf(stderr, " more than one \\special file name given: %s ignored.", sf );
            }
            (void)strcpy(spbuf, k.Key);
            sf = spbuf;
         } else
            if( GetKeyVal( &k, KeyTab, NKEYS, &j ) && j != -1 ) {
               if( j >= 0 && j <= 2 ) {
                  if( sf ) {
                     (void)fprintf(stderr, " more than one \\special file name given: %s ignored.", sf );
                  }
                  (void)strcpy(fbuf, k.Key);
                  (void)strcpy(spbuf, k.Val);
                  f = fbuf;
                  sf = spbuf;
                  if( j == 1 || j == 2)
                     ps = FALSE;
               } else /* keywords are output as PS procedure calls */
                  if(k.vt == Integer)
                     (void)fprintf(bitfile, "%ld @%s ", k.v.i, KeyTab[j].Entry);
                  else
                     (void)fprintf(bitfile, "%f @%s ", k.v.n, KeyTab[j].Entry);
            } else { 
               (void)fprintf(stderr, " invalid  keyword or value in \\special: \"%s\" ignored.", k.Key );
            }
      }

      cmdout("@setspecial\n");

      if( sf ) {
         figcopyfile(sf);
      } else
         (void)fprintf(stderr, " no special file name provided.");

      cmdout("@endspecial");
      }
   }
}

/*
 * extract first keyword-value pair from string (value part may be null)
 * return pointer to remainder of string return NULL if none found
 */
 
char    KeyStr[STRINGSIZE];
char    ValStr[STRINGSIZE];

char *GetKeyStr( str, kw )
char    *str;
KeyWord *kw;
{
   char *s, *k, *v, t;

   if( !str ) 
      return( NULL );

   for( s=str; *s == ' '; s++ )            /* skip over blanks */
      ;
   if( *s == '\0' ) 
      return( NULL );

   /* extract keyword portion */

   for( k=KeyStr; *s != ' ' && *s != '\0' && *s != '='; *k++ = *s++ )
      ;
   *k = '\0';
   kw->Key = KeyStr;
   kw->Val = v = NULL;
   kw->vt = None;

   for( ; *s == ' '; s++ )                 /* skip over blanks */
      ;
   if( *s != '=' )                         /* look for "=" */
      return( s );

   for( s++ ; *s == ' '; s++ )             /* skip over blanks */
      ;
   if( *s == '\'' || *s == '\"' )          /* get string delimiter */
      t = *s++;
   else
      t = ' ';

   /* copy value portion up to delimiter */

   for( v=ValStr; *s != t && *s != '\0'; *v++ = *s++ )
      ;
   if( t != ' ' && *s == t )
      s++;
   *v = '\0';
   kw->Val = ValStr;
   kw->vt = String;

   return( s );
}

/*
 * get next keyword-value pair decode value according to table entry
 */

int GetKeyVal( kw, tab, nt, tno)
KeyWord *kw; 
KeyDesc tab[];
int     nt;
int     *tno;
{
   int i;
   char c = '\0';

   *tno = -1;

   for(i=0; i<nt; i++)
      if( IsSame(kw->Key, tab[i].Entry) ) {
         *tno = i;
         switch( tab[i].Type ) {
            case None: 
               if( kw->vt != None ) 
                  return( 0 );
               break;
            case String:
               if( kw->vt != String ) 
                  return( 0 );
               break;
            case Integer:
               if( kw->vt != String ) 
                  return( 0 );
               if( sscanf(kw->Val,"%ld%c", &(kw->v.i), &c) != 1 || c != '\0' ) 
                  return( 0 );
               break;
            case Number:
            case Dimension:
               if( kw->vt != String ) 
                  return( 0 );
               if( sscanf(kw->Val,"%f%c", &(kw->v.n), &c) != 1 || c != '\0' ) 
                  return( 0 );
               break;
         }
         kw->vt = tab[i].Type;
         return( 1 );
      }

      return( 1 );
}

/*
 * compare strings, ignore case
 */
char Tolower(c)
register char c ;
{
   if ('A' <= c && c <= 'Z')
      return(c+32) ;
   else
      return(c) ;
}
int IsSame(a, b)
char *a, *b;
{
   for( ; *a != '\0'; )
      if( Tolower(*a++) != Tolower(*b++) ) 
         return( 0 );
      return( *a == *b ? 1 : 0 );
}
