/*
 *   This is dvips, a freely redistributable PostScript driver
 *   for dvi files.  It is (C) Copyright 1987 by Tomas Rokicki.
 *   You may modify and use this program to your heart's content,
 *   so long as you send modifications to Tomas Rokicki.  It can
 *   be included in any distribution, commercial or otherwise, so
 *   long as the banner string defined in structures.h is not
 *   modified (except for the version number) and this banner is
 *   printed on program invocation, or can be printed on program
 *   invocation with the -? option.
 */
#ifndef DEFRES
#define DEFRES (300)
#endif
#ifndef DEFPFMT
#define DEFPFMT "@letter"
#endif
/*
 *   Main routine for dvips.c.  (C) 1987 Radical Eye Software.
 */
#include "structures.h"
/*
 *   First we define some globals.
 */
fontdesctype *TeXfonts[256] ; /* font pointers */
fontdesctype *curfnt ;        /* the currently selected font */
sectiontype *sections ;       /* sections to process document in */
Boolean landscape ;           /* is this document in landscape? */
Boolean manualfeed ;          /* manual feed? */
Boolean compressed ;          /* compressed? */
int collatedcopies = 1 ;      /* how many collated copies? */
int sectioncopies = 1 ;       /* how many uncollated copies? */
shalfword linepos ;           /* where are we on the line? */
integer maxpages ;            /* the maximum number of pages */
Boolean notfirst ;            /* true if a first page was specified */
Boolean sendcontrolD ;        /* should we send a control D at end? */
integer firstpage ;           /* the number of the first page if specified */
Boolean orientationset ;      /* after this goes true, too late to landscape */
int numcopies ;          /* number of copies of this document to print */
char *oname ;                 /* output file name */
char *iname ;                 /* dvi file name */
char strings[STRINGSIZE] ;    /* strings for program */
char *nextstring, *maxstring ; /* string pointers */
FILE *dvifile, *bitfile ;     /* dvi and bit files */
FILE *pkfile ;                /* font file */
Boolean prescan ;             /* is this the prescan? */
integer curpos ;              /* dvi file location */
real conv ;                   /* conversion ratio from TeX units to pixels */
integer mag ;                 /* the magnification of this document */
int actualdpi = DEFRES ;      /* the actual resolution of the printer */
char *paperfmt = DEFPFMT ;    /* paper format */
integer fontmem = 800000 ;    /* memory remaining in printer */
integer pagecount ;           /* page counter for the sections */
integer pagenum ;             /* the page number we currently look at */
long bytesleft ;              /* number of bytes left in raster */
quarterword *raster ;         /* area for raster manipulations */
shalfword hh, vv ;            /* horizontal and vertical pixel positions */
shalfword fnt ;               /* current font */
char *tfmpath ;               /* pointer to directories for tfm files */
char *pkpath ;                /* pointer to directories for pk files */
integer swmem ;               /* memory in smartwriter */
int quiet ;                   /* should we only print errors to stderr? */
int filter ;                  /* act as filter default output to stdout,
                                               default input to stdin? */
int totalpages = 0 ;          /* total number of pages */
Boolean reverse ;             /* are we going reverse? */
Boolean usesPSfonts ;         /* do we use local PostScript fonts? */
Boolean headers_off ;         /* do we send headers or not? */
char *headerfile ;            /* default header file */
Boolean sendtrailer ;         /* send trailer?  (Are we EPSF or not?) */
Boolean disablecomments ;     /* should we send any EPSF comments? */
char *printer ;               /* what printer to send this to? */
char *mfmode ;                /* default MF mode */

#ifdef DEBUG
integer debug_flag = 0;
#endif /* DEBUG */
/*
 *   This routine calls the following externals:
 */
extern void outbangspecials() ;
extern void prescanpages() ;
extern void initprinter() ;
extern void cleanprinter() ;
extern void dosection() ;
extern void getdefaults() ;
extern void cmdout() ;
extern void numout() ;
extern void add_header() ;
extern void send_headers() ;
extern char *strcpy() ;
/*
 *   This error routine prints the error message; if the first
 *   character is !, it exits.
 */
static char *progname ;
void
error(s)
	char *s ;
{
   (void)fprintf(stderr, "%s: %s\n", progname, s) ;
   if (*s=='!') {
      if (bitfile != NULL) {
         cleanprinter() ;
      }
      exit(1) ;
   }
}
/*
 *   Initialize sets up all the globals and data structures.
 */
void
initialize()
{
   int i ;

   for (i=0; i<256; i++)
      TeXfonts[i] = NULL ;
   sections = NULL ;
   landscape = 0 ;
   maxpages = 100000 ;
   firstpage = 0 ;
   notfirst = 0 ;
   orientationset = 0 ;
   numcopies = 1 ;
   nextstring = strings ;
   oname = strings ;
   iname = strings ;
   *nextstring++ = 0 ;
   maxstring = strings + STRINGSIZE - 200 ;
   bitfile = NULL ;
   bytesleft = 0 ;
   pkpath = PKPATH ;
   tfmpath = TFMPATH ;
   swmem = SWMEM ;
   oname = OUTPATH ;
#ifdef DEBUG
   debug_flag = 0;
#endif /* DEBUG */
}
/*
 *   This routine copies a string into the string `pool', safely.
 */
char *
newstring(s)
   char *s ;
{
   int l ;

   if (s == NULL)
      return(NULL) ;
   l = strlen(s) ;
   if (nextstring + l >= maxstring)
      error("! out of string space") ;
   (void)strcpy(nextstring, s) ;
   s = nextstring ;
   nextstring += l + 1 ;
   return(s) ;
}
/*
 *   Finally, our main routine.
 */
main(argc, argv)
	int argc ;
	char *argv[] ;
{
   int i, lastext = -1 ;
   register sectiontype *sects ;

   progname = argv[0] ;
   sendcontrolD = 0 ;
   initialize() ;
   getdefaults() ;
/*
 *   This next whole big section of code is straightforward; we just scan
 *   the options.  Allowable options are:  n (number of pages), c (number
 *   of copies), p (first page), and o (output file name.)  The argument
 *   can either immediately follow the option letter or be separated by
 *   spaces.  Any argument not preceeded by '-' is considered a file name;
 *   the program complains if more than one is given, and uses stdin 
 *   if none is given.
 */
   for (i=1; i<argc; i++) {
      if (*argv[i]=='-') {
         char *p=argv[i]+2 ;
         char c=argv[i][1] ;
         switch (c) {
case 'c' :
            if (*p == 0 && argv[i+1])
               p = argv[++i] ;
            if (sscanf(p, "%d", &numcopies)==0)
               error("! bad number of copies option.") ;
            break ;
#ifdef DEBUG
case 'd' :
	    if (*p == 0 && argv[i+1])
	       p = argv[++i];
	    if (sscanf(p, "%d", &debug_flag)==0)
	       error("! bad debug option.");
	    break;
#endif /* DEBUG */
case 'f' :
            filter = (*p != '0') ;
            if (filter)
               oname = "" ;
            break ;
case 'h' : case 'H' :
            if (*p == 0 && argv[i+1])
               p = argv[++i] ;
            if (strcmp(p, "-") == 0)
               headers_off = 1 ;
            else
               add_header(p) ;
            break ;
case 'm' :
            manualfeed = (*p != '0') ;
            break ;
case 'n' :
            if (*p == 0 && argv[i+1])
               p = argv[++i] ;
#ifdef SHORTINT
            if (sscanf(p, "%ld", &maxpages)==0)
#else	/* ~SHORTINT */
            if (sscanf(p, "%d", &maxpages)==0)
#endif	/* ~SHORTINT */
               error("! bad number of pages option.") ;
            break ;
case 'o' : case 'O' :
            if (*p == 0 && argv[i+1])
               p = argv[++i] ;
            oname = p ;
            break ;
case 'p' :
            if (*p == 0 && argv[i+1])
               p = argv[++i] ;
#ifdef SHORTINT
            if (sscanf(p, "%ld", &firstpage)==0)
#else	/* ~SHORTINT */
            if (sscanf(p, "%d", &firstpage)==0)
#endif	/* ~SHORTINT */
               error("! bad first page option.") ;
            else notfirst = 1 ;
            break ;
case 'q' : case 'Q' :
            quiet = (*p != '0') ;
            break ;
case 'r' :
            reverse = (*p != '0') ;
            break ;
case 't' :
            if (*p == 0 && argv[i+1])
               p = argv[++i] ;
            paperfmt = p ;
            break ;
case 'C' :
            if (*p == 0 && argv[i+1])
               p = argv[++i] ;
            if (sscanf(p, "%d", &collatedcopies)==0)
               error("! bad number of collated copies option.") ;
            break ;
case 'D' :
            if (*p == 0 && argv[i+1])
               p = argv[++i] ;
            if (sscanf(p, "%d", &actualdpi)==0 || actualdpi < 10 ||
                       actualdpi > 6000)
               error("! bad dpi parameter") ;
            break ;
case 'F' :
            sendcontrolD = (*p != '0') ;
            break ;
case 'N' :
            disablecomments = (*p != '0') ;
            break ;
case 'P' :
            if (*p == 0 && argv[i+1])
               p = argv[++i] ;
            printer = p ;
            getdefaults() ;
            break ;
case 'R' :
            reverse = 0 ;
            break ;
case 'Z' :
            compressed = (*p != '0') ;
            break ;
case '?' :
            (void)fprintf(stderr, BANNER) ;
            break ;
default:
            error("! bad option, not one of cdfhmnopqrtCDFPRZ?") ;
         }
      } else {
         if (*iname == 0) {
            register char *p ;

            lastext = 0 ;
            iname = nextstring ;
            p = argv[i] ;
            while (*p) {
               *nextstring = *p++ ;
               if (*nextstring == '.')
                  lastext = nextstring - iname ;
               else if (*nextstring == '/' || *nextstring == ':')
                  lastext = 0 ;
               nextstring++ ;
            }
            if (lastext == 0) {
               lastext = nextstring - iname ;
               *nextstring++ = '.' ;
               *nextstring++ = 'd' ;
               *nextstring++ = 'v' ;
               *nextstring++ = 'i' ;
            }
            *nextstring++ = 0 ;
         } else
            error("! two input file names specified on input") ;
      }
   }
   if (!quiet)
      (void)fprintf(stderr, BANNER) ;
   if (*oname == 0 && ! filter) {
      oname = nextstring ;
      for (i=0; i<=lastext; i++)
         *nextstring++ = iname[i] ;
      *nextstring++ = 'p' ;
      *nextstring++ = 's' ;
      *nextstring++ = 0 ;
   }
#ifdef DEBUG
   if (dd(D_PATHS)) {
#ifdef SHORTINT
	(void)fprintf(stderr,"input file %s output file %s swmem %ld\n",
#else /* ~SHORTINT */
   	(void)fprintf(stderr,"input file %s output file %s swmem %d\n",
#endif /* ~SHORTINT */
           iname, oname, swmem) ;
   (void)fprintf(stderr,"tfm path %s pk path %s\n", tfmpath, pkpath) ;
   } /* dd(D_PATHS) */
#endif /* DEBUG */
/*
 *   Now we try to open the dvi file.
 */
   headerfile = (compressed? CHEADERFILE : HEADERFILE) ;
   add_header(headerfile) ;
   if (*iname != 0)
      dvifile = fopen(iname, "r") ;
   else if (filter)
      dvifile = stdin;
   else
      error("! no input filename supplied") ;
   if (dvifile==NULL)
      error("! couldn't open dvi file") ;
   curpos = 0 ;
/*
 *   Now we do our main work.
 */
   prescan = 1 ;
   prescanpages() ;
   if (usesPSfonts)
      add_header(PSFONTHEADER) ;
   sects = sections ;
   initprinter((sects == NULL || sects->next == NULL) && !disablecomments) ;
   outbangspecials() ;
   prescan = 0 ;
   if (sects == NULL || sects->next == NULL) {
      sectioncopies = collatedcopies ;
      collatedcopies = 1 ;
      sendtrailer = 1 ;
   } else {
      totalpages *= collatedcopies ;
   }
   for (i=0; i<collatedcopies; i++) {
      sects = sections ;
      while (sects != NULL) {
         if (! quiet)
            (void)fprintf(stderr, ". ") ;
         (void)fflush(stderr) ;
         dosection(sects, sectioncopies) ;
         sects = sects->next ;
      }
   }
   cleanprinter() ;
   if (! quiet)
      (void)fprintf(stderr, "\n") ;
   exit(0) ;
   /*NOTREACHED*/
}
