/*
 *   This is dvips, a freely redistributable PostScript driver
 *   for dvi files.  It is (C) Copyright 1987 by Tomas Rokicki.
 *   You may modify and use this program to your heart's content,
 *   so long as you send modifications to Tomas Rokicki.  It can
 *   be included in any distribution, commercial or otherwise, so
 *   long as the banner string defined in structures.h is not
 *   modified (except for the version number) and this banner is
 *   printed on program invocation, or can be printed on program
 *   invocation with the -? option.
 */
/*
 *   output.c of dvisw software package.  This code (C) 1986 Radical Eye
 *   Software.
 *
 *   These routines do most of the communicating with the printer.
 *
 *   LINELENGTH tells the maximum line length to send to the printer.
 */
#define LINELENGTH (78)
#include "structures.h"
#include <ctype.h>
/*
 *   The external routines called here:
 */
extern void error() ;
extern void send_headers() ;
extern FILE *search() ;
/*
 *   These are the external variables used by these routines.
 */
extern shalfword hh, vv ;
extern fontdesctype *curfnt ;
extern FILE *bitfile ;
extern char *oname ;
extern Boolean landscape, reverse ;
extern Boolean sendcontrolD ;
extern Boolean usesPSfonts, headers_off ;
extern int numcopies ;
extern int actualdpi ;
extern int totalpages ;
extern integer pagenum ;
extern Boolean manualfeed ;
extern int quiet ;
extern char *iname ;
extern char *paperfmt ;
extern char *getenv() ;
/*
 *   We need a few statics to take care of things.
 */
static shalfword rhh, rvv ;
shalfword linepos = 0 ;
static Boolean instring ;
static Boolean lastspecial = 1 ;
static shalfword d ;
static int lastfont ;
static void chrcmd();			/* just a forward declaration */

/*
 *   Some very low level primitives to send output to the printer.
 */

/*
 *   This routine copies a file down the pipe.  Search path uses the
 *   header path.
 */
int epsfc ;
static int infigure ;
void
copyfile(s)
	char *s ;
{
   FILE *f = search(HEADERPATH, s) ;
   int c, prevc = 10 ;

   if (f==NULL) {
      if (infigure) {
         (void)fprintf(stderr, "(%s) %s\n", HEADERPATH, s) ;
         error(" couldn't find figure file; continuing") ;
      } else {
         (void)fprintf(stderr, "(%s) %s\n", HEADERPATH, s) ;
         error("! couldn't find header file") ;
      }
   } else {
      if (! quiet) {
         (void)fprintf(stderr, "[%s]", s) ;
         (void)fflush(stderr) ;
      }
      if (linepos != 0)
         (void)putc('\n', bitfile) ;
      if (epsfc)
         (void)fprintf(bitfile, "%%%%BeginDocument: %s\n", s) ;
      while ((c=getc(f))!=EOF) {
         if (c == '%' && prevc == '\n') { /* skip comments */
            while ((c=getc(f))!=EOF) {
               if (c=='\n')
                  break ;
            }
         } else
            (void)putc(c, bitfile) ;
         prevc = c ;
      }
      if (prevc != '\n')
         (void)putc('\n', bitfile) ;
      linepos = 0 ;
      (void)fclose(f) ;
      if (epsfc)
         (void)fprintf(bitfile, "%%%%EndDocument\n") ;
   }
}

/*
 *   For included PostScript graphics, we use the above routine, but
 *   with no fatal error message.
 */
void figcopyfile(s)
char *s ;
{
   infigure = 1 ;
   copyfile(s) ;
   infigure = 0 ;
}

/*
 *   This next routine writes out a `special' character.  In this case,
 *   we simply put it out, since any special character terminates the
 *   preceding token.
 */
void
specialout(c)
	char c ;
{
   if (linepos + 1 > LINELENGTH) {
      (void)putc('\n', bitfile) ;
      linepos = 0 ;
   }
   (void)putc(c, bitfile) ;
   linepos++ ;
   lastspecial = 1 ;
}

static void
scout(c)
	char c ;
{
   if (! instring) {
/*
 *   If we are within 5, we send a carriage return.
 *   This eliminates extraneous backslash-newlines.
 */
      if (linepos + 5 > LINELENGTH) {
         (void)putc('\n', bitfile) ;
         linepos = 0 ;
      }
      (void)putc('(', bitfile) ;
      linepos++ ;
      lastspecial = 1 ;
      instring = 1 ;
   }
   if (c<' ' || c>126) {
      if (linepos + 5 > LINELENGTH) {
         (void)putc('\\', bitfile) ;
         (void)putc('\n', bitfile) ;
         linepos = 0 ;
      }
      (void)putc('\\', bitfile) ;
      linepos++ ;
      (void)putc('0' + ((c >> 6) & 7), bitfile) ;
      linepos++ ;
      (void)putc('0' + ((c >> 3) & 7), bitfile) ;
      linepos++ ;
      (void)putc('0' + (c & 7), bitfile) ;
      linepos++ ;
   } else if (c == '(' || c == ')' || c == '\\') {
      if (linepos + 3 > LINELENGTH) {
         (void)putc('\\', bitfile) ;
         (void)putc('\n', bitfile) ;
         linepos = 0 ;
      }
      (void)putc('\\', bitfile) ;
      linepos++ ;
      (void)putc(c, bitfile) ;
      linepos++ ;
   } else {
      if (linepos + 2 > LINELENGTH) {
         (void)putc('\\', bitfile) ;
         (void)putc('\n', bitfile) ;
         linepos = 0 ;
      }
      (void)putc(c, bitfile) ;
      linepos++ ;
   }
}

static void
stringend()
{
   instring = 0 ;
   specialout(')') ;
}

void
cmdout(s)
	char *s ;
{
   int l ;

   /* hack added by dorab -- dont know if this is right */
   if (instring) {
	stringend();
	chrcmd('p');
   }
   l = strlen(s) ;
   if (linepos + l + 1 > LINELENGTH) {
      (void)putc('\n', bitfile) ;
      linepos = 0 ;
      lastspecial = 1 ;
   }
   if (! lastspecial) {
      (void)putc(' ', bitfile) ;
      linepos++ ;
   }
   while (*s != 0) {
      (void)putc(*s++, bitfile) ;
   }
   linepos += l ;
   lastspecial = 0 ;
}

static void
chrcmd(c)
	char c ;
{
   if (linepos + 2 > LINELENGTH) {
      (void)putc('\n', bitfile) ;
      linepos = 0 ;
      lastspecial = 1 ;
   }
   if (! lastspecial) {
      (void)putc(' ', bitfile) ;
      linepos++ ;
   }
   (void)putc(c, bitfile) ;
   linepos++ ;
   lastspecial = 0 ;
}

void
numout(n)
	integer n ;
{
   char buf[10] ;

   (void)sprintf(buf, "%ld", n) ;
   cmdout(buf) ;
}

void mhexout(p, len)
register unsigned char *p ;
register long len ;
{
   register char *hexchar = "0123456789ABCDEF" ;
   register int linep ;
   register int n ;

   linep = linepos ;
   while (len > 0) {
      if (linep > LINELENGTH - 2) {
         (void)putc('\n', bitfile) ;
         linep = 0 ;
      }
      n = *p++ ;
      (void)putc(hexchar[n >> 4], bitfile) ;
      (void)putc(hexchar[n & 15], bitfile) ;
      linep += 2 ;
      len-- ;
   }
   linepos = linep ;
}

void
fontout(n)
	int n ;
{
   char buf[6] ;

   if (instring) {
      stringend() ;
      chrcmd('p') ;
   }
   if (n < 27)
      (void)sprintf(buf, "f%c", 'a'+n-1) ;
   else
      (void)sprintf(buf, "f%d", n-27) ;
   cmdout(buf) ;
}

void
hvpos()
{
   int is = instring ;

   if (rvv != vv) {
      if (instring)
         stringend() ;
      numout((integer)hh) ;
      numout((integer)vv) ;
      chrcmd(is ? 'y' : 'a') ;
      rvv = vv ;
   } else if (rhh != hh) {
      if (instring) {
         stringend() ;
         if (hh - rhh < 5 && rhh - hh < 5) {
            chrcmd('p' + hh - rhh) ;
         } else if (hh - rhh < d + 5 && rhh - hh < 5 - d) {
            chrcmd('g' + hh - rhh - d) ;
            d = hh - rhh ;
         } else {
            numout((integer)hh - rhh) ;
            chrcmd('b') ;
            d = hh - rhh ;
         }
      } else {
         numout((integer)hh - rhh) ;
         chrcmd('w') ;
      }
   }
   rhh = hh ;
}

/*
 *   initprinter opens the bitfile and writes the initialization sequence
 *   to it.
 */
void newline()
{
   if (linepos != 0) {
      (void)fprintf(bitfile, "\n") ;
      linepos = 0 ;
   }
}
void
initprinter(epsf)
int epsf ;
{
   epsfc = epsf ;
   if (*oname != 0) {
/*
 *   We check to see if the first character is a exclamation
 *   point, and popen if so.
 */
      if (*oname == '!' || *oname == '|') {
         if ((bitfile=popen(oname+1, "w"))==NULL)
            error("! couldn't open output pipe") ;
      } else {
         if ((bitfile=fopen(oname,"w"))==NULL)
            error("! couldn't open PostScript file") ;
      }
   } else {
      bitfile = stdout ;
   }
   if (epsf)
      (void)fprintf(bitfile, "%%!PS-Adobe-2.0\n") ;
   else
      (void)fprintf(bitfile, "%%!PS (but not EPSF through switch or no memory)\n") ;
   (void)fprintf(bitfile, "%%%%Creator: dvips by Radical Eye Software\n") ;
   if (*iname)
      (void)fprintf(bitfile, "%%%%Title: %s\n", iname) ;
   (void)fprintf(bitfile, "%%%%Pages: %d %d\n", totalpages, 1 - 2 * reverse) ;
   if (landscape)
      (void)fprintf(bitfile, "%%%%BoundingBox: 0 0 792 612\n") ;
   else
      (void)fprintf(bitfile, "%%%%BoundingBox: 0 0 612 792\n") ;
   if (epsfc)
      (void)fprintf(bitfile, "%%%%EndComments\n") ;
   if (! headers_off)
      send_headers() ;
}
static int endprologsent ;
setup() {
   cmdout("end") ;
   newline() ;
   if (endprologsent == 0 && epsfc) {
      (void)fprintf(bitfile, "%%%%EndProlog\n") ;
      (void)fprintf(bitfile, "%%%%BeginSetup\n") ;
      (void)fprintf(bitfile, "%%%%Feature: *Resolution %d\n", DPI) ;
   }
   cmdout("TeXDict") ;
   cmdout("begin") ;
   if (paperfmt[0] != '@') {
      char pft[100] ;
      strcpy(pft, "@") ;
      strcat(pft, paperfmt) ;
      cmdout(pft) ;
   } else
      cmdout(paperfmt) ;
   if (manualfeed) cmdout("@manualfeed") ;
   if (numcopies != 1) {
      numout((integer)numcopies) ;
      cmdout("@copies") ;
   }
   if (endprologsent == 0 && epsfc) {
      newline() ;
      endprologsent = 1 ;
      (void)fprintf(bitfile, "%%%%EndSetup\n") ;
   }

}
/*
 *   cleanprinter is the antithesis of the above routine.
 */
void
cleanprinter()
{
   (void)fprintf(bitfile, "\n") ;
   (void)fprintf(bitfile, "%%%%EOF\n") ;
   if (sendcontrolD)
      (void)putc(4, bitfile) ;
   (void)fclose(bitfile) ;
   bitfile = NULL ;
}

/*
 *   pageinit initializes the output variables.
 */
static int thispage = 0 ;
void
pageinit()
{
   rhh = -3200 ;
   rvv = -3200 ;
   newline() ;
   (void)fprintf(bitfile, "%%%%Page: %ld %d\n", pagenum, ++thispage) ;
   linepos = 0 ;
   cmdout("bop") ;
   lastfont = -1 ;
   d = -3200 ;
}

/*
 *   This routine ends a page.
 */
void
pageend()
{
   if (instring) {
      stringend() ;
      chrcmd('g') ;
   }
   cmdout("eop") ;
}

/*
 *   drawrule draws a rule at the specified position.
 */
void
drawrule(rw, rh)
	shalfword rw, rh ;
{
   hvpos() ;
   if (instring) {
      stringend() ;
      chrcmd('p') ;
   }
   numout((integer)rw) ;
   numout((integer)rh) ;
   chrcmd('v') ;
}

/*
 *   drawchar draws a character at the specified position.
 */
void
drawchar(c, cc)
	chardesctype *c ;
	int cc ;
{
   hvpos() ;
   if (lastfont != curfnt->psname) {
      fontout((int)curfnt->psname) ;
      lastfont = curfnt->psname ;
   }
   scout(cc) ;
   rhh = hh + c->pixelwidth ;
   rvv = vv ;
}
