/*
 *   This is dvips, a freely redistributable PostScript driver
 *   for dvi files.  It is (C) Copyright 1987 by Tomas Rokicki.
 *   You may modify and use this program to your heart's content,
 *   so long as you send modifications to Tomas Rokicki.  It can
 *   be included in any distribution, commercial or otherwise, so
 *   long as the banner string defined in structures.h is not
 *   modified (except for the version number) and this banner is
 *   printed on program invocation, or can be printed on program
 *   invocation with the -? option.
 */
/*
 *   resident.c of dvisw software package.  (C) 1986 Radical Eye Software.
 *   This code reads in and handles the defaults for the program from the
 *   file config.sw.  This entire file is a bit kludgy, sorry.
 */
#include "structures.h"
/*
 *   This is the structure definition for resident fonts.  We use
 *   a small and simple hash table to handle these.  We don't need
 *   a big hash table.
 */
struct resfont *reshash[RESHASHPRIME] ;
/*
 *   These are the external routines we use.
 */
extern void error() ;
extern integer scalewidth() ;
extern void tfmload() ;
extern FILE *search() ;
extern shalfword pkbyte() ;
extern integer pkquad() ;
extern integer pktrio() ;
extern Boolean pkopen() ;
extern char *strcpy() ;
extern char *newstring() ;
extern void add_header() ;
/*
 *   These are the external variables we use.
 */
#ifdef DEBUG
extern integer debug_flag;
#endif	/* DEBUG */
extern fontdesctype *TeXfonts[] ;
extern integer mag ;
extern long bytesleft ;
extern quarterword *raster ;
extern FILE *pkfile ;
extern char *oname ;
extern integer swmem ;
extern char *tfmpath ;
extern char *pkpath ;
extern char *nextstring ;
extern char *maxstring ;
extern Boolean landscape ;
extern Boolean disablecomments ;
extern int quiet ;
extern int filter ;
extern Boolean reverse ;
extern Boolean usesPSfonts ;
extern int actualdpi ;
extern char *printer ;
extern char *mfmode ;
/*
 *   We use malloc here.
 */
char *malloc() ;
/*
 *   Our hash routine.
 */
int
hash(s)
   char *s ;
{
   int h = 12 ;

   while (*s != 0)
      h = (h + h + *s++) % RESHASHPRIME ;
   return(h) ;
}
/*
 *   Cleanres marks all resident fonts as not being yet sent.
 */
cleanres() {
   register int i ;
   register struct resfont *p ;

   for (i=0; i<RESHASHPRIME; i++)
      for (p=reshash[i]; p; p=p->next)
         p->sent = 0 ;
}
/*
 *   The routine which looks up a font name.
 */
struct resfont *
lookup(name)
   char *name ;
{
   struct resfont *p ;

   for (p=reshash[hash(name)]; p!=NULL; p=p->next)
      if (strcmp(p->TeXname, name)==0)
         return(p) ;
   return(NULL) ;
}
/*
 *   This routine adds an entry.
 */
void
add_entry(TeXname, PSname, specinfo)
   char *TeXname, *PSname, *specinfo ;
{
   struct resfont *p ;
   int h ;

   if (PSname == NULL)
      PSname = TeXname ;
   else if (PSname != TeXname)
      add_entry(PSname, PSname, specinfo) ;
   p = (struct resfont *)malloc((unsigned int)sizeof(struct resfont)) ;
   if (p==NULL)
      error("! out of memory") ;
   p->TeXname = TeXname ;
   p->PSname = PSname ;
   p->specialinstructions = specinfo ;
   h = hash(TeXname) ;
   p->next = reshash[h] ;
   reshash[h] = p ;
}
/*
 *   Now our residentfont routine.
 */
Boolean
residentfont(fn)
	halfword fn ;
{
   register shalfword i ;
   register fontdesctype *curfnt ;
   char *nm ;
   struct resfont *p ;

/*
 *   First we determine if we can find this font in the resident list.
 */
   for (nm=TeXfonts[fn]->name; *nm != 0; nm++) ;
   nm++ ;
   if ((p=lookup(nm))==NULL)
      return 0 ;
/*
 *   We clear out some pointers:
 */
#ifdef DEBUG
   if (dd(D_FONTS))
	(void)fprintf(stderr,"This font is resident.\n") ;
#endif	/* DEBUG */
   curfnt = TeXfonts[fn] ;
   curfnt->resfont = p ;
   for (i=0; i<256; i++) {
      curfnt->chardesc[i].TFMwidth = 0 ;
      curfnt->chardesc[i].packptr = NULL ;
      curfnt->chardesc[i].pixelwidth = 0 ;
      curfnt->chardesc[i].flags = EXISTS ;
   }
   tfmload((halfword)fn) ;
   usesPSfonts = 1 ;
   return(1) ;
}
static char was_inline[100] ;
void
bad_config() {
   error("Error in config file:") ;
   error(was_inline) ;
}
/*
 *   Now we have the getdefaults routine.
 */
void
getdefaults() {
   FILE *deffile ;
   char PSname[100] ;
   register char *p ;
   char *specinfo, *PS, *fontname ;

   if (printer == NULL)
      strcpy(PSname, CONFIGFILE) ;
   else {
      strcpy(PSname, "config.") ;
      strcat(PSname, printer) ;
   }
   if ((deffile=search(CONFIGPATH,PSname))!=NULL) {
      while (fgets(was_inline, 100, deffile)!=NULL) {
       p = was_inline + strlen(was_inline) - 1 ;
       if (p >= was_inline && *p == '\n')
          *p = 0 ;
       switch (was_inline[0]) {
case 'm' :
#ifdef SHORTINT
         if (sscanf(was_inline+1, "%ld", &swmem) != 1) bad_config() ;
#else	/* ~SHORTINT */
         if (sscanf(was_inline+1, "%d", &swmem) != 1) bad_config() ;
#endif	/* ~SHORTINT */
         break ;
case 'M' :
         if (sscanf(was_inline+1, "%s", PSname) != 1) bad_config() ;
         mfmode = newstring(PSname) ;
         break ;
case 'o' : case 'O' :
         p = was_inline + 1 ;
         while (*p && *p <= ' ')
            p++ ;
         oname = newstring(p) ;
         break ;
case 't' : case 'T' :
         if (sscanf(was_inline+1, "%s", PSname) != 1) bad_config() ;
         else tfmpath = newstring(PSname) ;
         break ;
case 'p' : case 'P' :
         if (sscanf(was_inline+1, "%s", PSname) != 1) bad_config() ;
         else pkpath = newstring(PSname) ;
         break ;
case ' ' : case '*' : case '#' : case ';' : case '=' : case 0 :
         break ;
case 'r' : case 'R' :
         reverse = 1 ;
         break ;
case 'D' :
         if (sscanf(was_inline+1, "%d", &actualdpi) != 1) bad_config() ;
         if (actualdpi < 50 || actualdpi > 10000) bad_config() ;
         break ;
case 'q' : case 'Q' :
         quiet = 1 ;
         break ;
case 'f' : case 'F' :
         filter = 1 ;
         break ;
case 'h' : case 'H' :
         if (sscanf(was_inline+1, "%s", PSname) != 1) bad_config() ;
         else add_header(PSname) ;
         break ;
case 'N' :
         disablecomments = 1 ;
         break ;
default:
         bad_config() ;
      }
     }
     (void)fclose(deffile) ;
     if (printer)
        return ;
   }
   if ((deffile=search(CONFIGPATH,PSMAPFILE))!=NULL) {
     while (fgets(was_inline, 100, deffile)!=NULL) {
         specinfo = NULL ;
         fontname = NULL ;
         PS = NULL ;
         p = was_inline ;
         while (*p && *p <= ' ')
            p++ ;
         if (*p) fontname = p ;
         while (*p > ' ')
            p++ ;
         if (*p)
            *p++ = 0 ;
         while (*p && *p <= ' ')
            p++ ;
         if (*p != '"' && *p) {
            PS = p ;
            while (*p > ' ')
               p++ ;
            if (*p)
               *p++ = 0 ;
            while (*p && *p <= ' ')
               p++ ;
         }
         if (*p == '"') {
            p++ ;
            specinfo = p ;
            while (*p && *p != '"')
               p++ ;
            *p = 0 ;
         }
         add_entry(newstring(fontname), newstring(PS), newstring(specinfo)) ;
       }
       (void)fclose(deffile) ;
   }
   checkenv() ;
}
/*
 *   Get environment variables!
 */
checkenv() {
   char *p, *getenv() ;

   if (p=getenv("TEXFONTS"))
      tfmpath = newstring(p) ;
   if (p=getenv("TEXPKS"))
      pkpath = newstring(p) ;
}
